@@ -24,20 +24,22 @@ fix_word_t fw_remainder = 0;
 
 double fw2double(fix_word_t fw)
 {
-        return ((double) fw) / (1 << FW_FRACTION_BIT);
+        return ((double) fw) / (((uint32) 1) << FW_FRACTION_BIT);
 }
 
+
 fix_word_t fw_prod(fix_word_t a, fix_word_t b)
 {
-        fix_word_t al, bl ;
-        
-        if (a < 0)
-                return fw_prod(-a, b) ;
-        if (b < 0)
-                return fw_prod(a, -b) ;
-        al = a & 32767 ;
-        bl = b & 32767 ;
-        a >>= 15 ;
-        b >>= 15 ;
-        return ( ((al*bl/32768) + a*bl+al*b)/32 + a*b*1024) ;
+        fix_word_t al, bl;
+
+        if (a < 0) return - fw_prod(-a, b);
+        if (b < 0) return - fw_prod(a, -b);
+	
+	/* We have to compute a*b >> 20 without overflowing 32 bits */
+        al = a & 32767;
+        bl = b & 32767;
+        a >>= 15;
+        b >>= 15;
+	
+	return (((al*bl >> 15) + a*bl + al*b) >> 5) + (a*b << 10);
 }
