@@ -145,8 +145,8 @@ static const glyph_t genx_2bracketleft[] = {GLYPH_bracketleft, 0};
 static const glyph_t genx_2bracketright[] = {GLYPH_bracketright, 0};
 static const glyph_t genx_2braceleft[] = {GLYPH_braceleft, 0};
 static const glyph_t genx_2braceright[] = {GLYPH_braceright, 0};
-static const glyph_t genx_2angleleft[] = {GLYPH_angleleft, 0};
-static const glyph_t genx_2angleright[] = {GLYPH_angleright, 0};
+static const glyph_t genx_2angleleftmath[] = {GLYPH_UNI_angleleftmath, 0};
+static const glyph_t genx_2anglerightmath[] = {GLYPH_UNI_anglerightmath, 0};
 static const glyph_t genx_2UNI_floorleft[] = {GLYPH_UNI_floorleft, 0};
 static const glyph_t genx_2UNI_floorright[] = {GLYPH_UNI_floorright, 0};
 static const glyph_t genx_2UNI_ceilingleft[] = {GLYPH_UNI_ceilingleft, 0};
@@ -216,14 +216,14 @@ static const glyphtweak_t generic_glyphtweaks[] = {
     {GLYPH_CATDVI_bracerightBig,    genx_2braceright,	    WIDTH_AUTO},
     {GLYPH_CATDVI_bracerightbigg,   genx_2braceright,	    WIDTH_AUTO},
     {GLYPH_CATDVI_bracerightBigg,   genx_2braceright,	    WIDTH_AUTO},
-    {GLYPH_CATDVI_angbracketleftbig, genx_2angleleft,	    WIDTH_AUTO},
-    {GLYPH_CATDVI_angbracketleftBig, genx_2angleleft,	    WIDTH_AUTO},
-    {GLYPH_CATDVI_angbracketleftbigg, genx_2angleleft,	    WIDTH_AUTO},
-    {GLYPH_CATDVI_angbracketleftBigg, genx_2angleleft,	    WIDTH_AUTO},
-    {GLYPH_CATDVI_angbracketrightbig, genx_2angleright,	    WIDTH_AUTO},
-    {GLYPH_CATDVI_angbracketrightBig, genx_2angleright,	    WIDTH_AUTO},
-    {GLYPH_CATDVI_angbracketrightbigg, genx_2angleright,    WIDTH_AUTO},
-    {GLYPH_CATDVI_angbracketrightBigg, genx_2angleright,    WIDTH_AUTO},
+    {GLYPH_CATDVI_angbracketleftbig, genx_2angleleftmath,   WIDTH_AUTO},
+    {GLYPH_CATDVI_angbracketleftBig, genx_2angleleftmath,   WIDTH_AUTO},
+    {GLYPH_CATDVI_angbracketleftbigg, genx_2angleleftmath,  WIDTH_AUTO},
+    {GLYPH_CATDVI_angbracketleftBigg, genx_2angleleftmath,  WIDTH_AUTO},
+    {GLYPH_CATDVI_angbracketrightbig, genx_2anglerightmath, WIDTH_AUTO},
+    {GLYPH_CATDVI_angbracketrightBig, genx_2anglerightmath, WIDTH_AUTO},
+    {GLYPH_CATDVI_angbracketrightbigg, genx_2anglerightmath,	WIDTH_AUTO},
+    {GLYPH_CATDVI_angbracketrightBigg, genx_2anglerightmath,    WIDTH_AUTO},
     {GLYPH_CATDVI_floorleftbig,     genx_2UNI_floorleft,    WIDTH_AUTO},
     {GLYPH_CATDVI_floorleftBig,     genx_2UNI_floorleft,    WIDTH_AUTO},
     {GLYPH_CATDVI_floorleftbigg,    genx_2UNI_floorleft,    WIDTH_AUTO},
@@ -289,7 +289,11 @@ static const glyphtweak_t generic_glyphtweaks[] = {
     {GLYPH_CATDVI_hatwidest,	genx_2asciicircum,	    WIDTH_AUTO},
     {GLYPH_CATDVI_tildewide,	genx_2asciitilde,	    WIDTH_AUTO},
     {GLYPH_CATDVI_tildewider,	genx_2asciitilde,	    WIDTH_AUTO},
-    {GLYPH_CATDVI_tildewidest,	genx_2asciitilde,	    WIDTH_AUTO}
+    {GLYPH_CATDVI_tildewidest,	genx_2asciitilde,	    WIDTH_AUTO},
+    /* End TeX Math Extension */
+
+    {GLYPH_CATDVI_hatxwide,	genx_2asciicircum,	    WIDTH_AUTO},
+    {GLYPH_CATDVI_tildexwide,	genx_2asciitilde,	    WIDTH_AUTO}
 };
 
 /* How to combine a base glyph and a single diacritic */
@@ -432,7 +436,7 @@ struct simple_charmap_entry_t {
     unsigned char   output;
 };
 
-static sparp_t simple_charmap_sparp;
+static spars32_t simple_charmap_spars32;
 static unsigned char simple_identity_range;
 
 /* implements outenc_write_raw_glyph for simple output encodings */
@@ -440,10 +444,10 @@ static void simple_write_raw_glyph(FILE * f, glyph_t g)
 {
     if(g <= simple_identity_range) putc((int) g, f);
     else {
-    	unsigned char const * p;
-	p = sparp_read(&simple_charmap_sparp, g);
+    	int c;
+	c = spars32_read(&simple_charmap_spars32, g);
 
-	if (p != NULL) putc(*p, f);
+	if (c > 0) putc(c, f);
 	else outenc_show_unknown(f, g);
     }
 }
@@ -465,15 +469,12 @@ static void simple_init_charmap(
     size_t i;
 
     simple_identity_range = identity_range;
-    sparp_init(&simple_charmap_sparp);
+    spars32_init(&simple_charmap_spars32, 0);
     for(i = 0; i < tbl_length; ++i) {
-    	/* The charmap sparp is indexed by the unicode value and points
-	 * directly to output.
-	 */
-    	sparp_write(
-	    &simple_charmap_sparp,
+    	spars32_write(
+	    &simple_charmap_spars32,
 	    charmap_tbl[i].unicode,
-	    &(charmap_tbl[i].output)
+	    charmap_tbl[i].output
 	);
     }
 
@@ -484,6 +485,40 @@ static void simple_init_charmap(
  * utf-8
  ***********************************************************************/
 
+/* A few characters in AMSSymbolsB are not in Unicode, but can be represented
+ * in Unicode by use of combining characters.
+ * FIXME: these translations should really be generic, but our substitution
+ * mechanism is not up to it.
+ */
+static const glyph_t utf8x_2not_lessequalslanted[] = {
+    GLYPH_UNI_lessequalslanted, GLYPH_UNI_slashlongcomb, 0
+};
+static const glyph_t utf8x_2not_greaterequalslanted[] = {
+    GLYPH_UNI_greaterequalslanted, GLYPH_UNI_slashlongcomb, 0
+};
+static const glyph_t utf8x_2not_lessequal2[] = {
+    GLYPH_UNI_lessequal2, GLYPH_UNI_slashlongcomb, 0
+};
+static const glyph_t utf8x_2not_greaterequal2[] = {
+    GLYPH_UNI_greaterequal2, GLYPH_UNI_slashlongcomb, 0
+};
+static const glyph_t utf8x_2not_subsetequal2[] = {
+    GLYPH_UNI_subsetequal2, GLYPH_UNI_slashlongcomb, 0
+};
+static const glyph_t utf8x_2not_supersetequal2[] = {
+    GLYPH_UNI_supersetequal2, GLYPH_UNI_slashlongcomb, 0
+};
+
+static const glyphtweak_t utf8_glyphtweaks[] = {
+    {GLYPH_CATDVI_notlessequalslanted, utf8x_2not_lessequalslanted, 1},
+    {GLYPH_CATDVI_notgreaterequalslanted, utf8x_2not_greaterequalslanted, 1},
+    {GLYPH_CATDVI_notlessequal2, utf8x_2not_lessequal2, 1},
+    {GLYPH_CATDVI_notgreaterequal2, utf8x_2not_greaterequal2, 1},
+    {GLYPH_CATDVI_notsubsetequal2, utf8x_2not_subsetequal2, 1},
+    {GLYPH_CATDVI_notsupersetequal2, utf8x_2not_supersetequal2, 1}
+};
+
+
 /* UCS-4 to UTF-8 translation control tables. */
 
 #define UTFVEC_MAXLEN (6+1) /* UTF-8 takes max 6 bytes + null terminator */
@@ -624,6 +659,10 @@ static void utf8_xlat_composition(
 
 static void utf8_init(void)
 {
+    outenc_register_glyphtweaks(
+    	utf8_glyphtweaks,
+	lengthof(utf8_glyphtweaks)
+    );
     outenc_write_raw_glyph = utf8_write_raw_glyph;
     outenc_composition_width = utf8_composition_width;
     outenc_xlat_composition = utf8_xlat_composition;
@@ -658,6 +697,74 @@ static const glyph_t a2l1x_UNI_longst[] = {GLYPH_s, GLYPH_t, 0};
 static const glyph_t a2l1x_minus[] = {GLYPH_UNI_hyphen, 0};
 static const glyph_t a2l1x_asteriskmath[] = {GLYPH_asterisk, 0};
 static const glyph_t a2l1x_2emptystring[] = {0};
+static const glyph_t a2l1x_2asciitilde[] = {GLYPH_asciitilde, 0};
+static const glyph_t a2l1x_lessequal[] = {GLYPH_equal, GLYPH_less, 0};
+static const glyph_t a2l1x_greaterequal[] = {GLYPH_greater, GLYPH_equal, 0};
+
+static const glyph_t a2l1x_2A[] = {GLYPH_A, 0};
+static const glyph_t a2l1x_2B[] = {GLYPH_B, 0};
+static const glyph_t a2l1x_2C[] = {GLYPH_C, 0};
+static const glyph_t a2l1x_2D[] = {GLYPH_D, 0};
+static const glyph_t a2l1x_2E[] = {GLYPH_E, 0};
+static const glyph_t a2l1x_2F[] = {GLYPH_F, 0};
+static const glyph_t a2l1x_2G[] = {GLYPH_G, 0};
+static const glyph_t a2l1x_2H[] = {GLYPH_H, 0};
+static const glyph_t a2l1x_2I[] = {GLYPH_I, 0};
+static const glyph_t a2l1x_2J[] = {GLYPH_J, 0};
+static const glyph_t a2l1x_2K[] = {GLYPH_K, 0};
+static const glyph_t a2l1x_2L[] = {GLYPH_L, 0};
+static const glyph_t a2l1x_2M[] = {GLYPH_M, 0};
+static const glyph_t a2l1x_2N[] = {GLYPH_N, 0};
+static const glyph_t a2l1x_2O[] = {GLYPH_O, 0};
+static const glyph_t a2l1x_2P[] = {GLYPH_P, 0};
+static const glyph_t a2l1x_2Q[] = {GLYPH_Q, 0};
+static const glyph_t a2l1x_2R[] = {GLYPH_R, 0};
+static const glyph_t a2l1x_2S[] = {GLYPH_S, 0};
+static const glyph_t a2l1x_2T[] = {GLYPH_T, 0};
+static const glyph_t a2l1x_2U[] = {GLYPH_U, 0};
+static const glyph_t a2l1x_2V[] = {GLYPH_V, 0};
+static const glyph_t a2l1x_2W[] = {GLYPH_W, 0};
+static const glyph_t a2l1x_2X[] = {GLYPH_X, 0};
+static const glyph_t a2l1x_2Y[] = {GLYPH_Y, 0};
+static const glyph_t a2l1x_2Z[] = {GLYPH_Z, 0};
+
+static const glyph_t a2l1x_2a[] = {GLYPH_a, 0};
+static const glyph_t a2l1x_2b[] = {GLYPH_b, 0};
+static const glyph_t a2l1x_2c[] = {GLYPH_c, 0};
+static const glyph_t a2l1x_2d[] = {GLYPH_d, 0};
+static const glyph_t a2l1x_2e[] = {GLYPH_e, 0};
+static const glyph_t a2l1x_2f[] = {GLYPH_f, 0};
+static const glyph_t a2l1x_2g[] = {GLYPH_g, 0};
+static const glyph_t a2l1x_2h[] = {GLYPH_h, 0};
+static const glyph_t a2l1x_2i[] = {GLYPH_i, 0};
+static const glyph_t a2l1x_2j[] = {GLYPH_j, 0};
+static const glyph_t a2l1x_2k[] = {GLYPH_k, 0};
+static const glyph_t a2l1x_2l[] = {GLYPH_l, 0};
+static const glyph_t a2l1x_2m[] = {GLYPH_m, 0};
+static const glyph_t a2l1x_2n[] = {GLYPH_n, 0};
+static const glyph_t a2l1x_2o[] = {GLYPH_o, 0};
+static const glyph_t a2l1x_2p[] = {GLYPH_p, 0};
+static const glyph_t a2l1x_2q[] = {GLYPH_q, 0};
+static const glyph_t a2l1x_2r[] = {GLYPH_r, 0};
+static const glyph_t a2l1x_2s[] = {GLYPH_s, 0};
+static const glyph_t a2l1x_2t[] = {GLYPH_t, 0};
+static const glyph_t a2l1x_2u[] = {GLYPH_u, 0};
+static const glyph_t a2l1x_2v[] = {GLYPH_v, 0};
+static const glyph_t a2l1x_2w[] = {GLYPH_w, 0};
+static const glyph_t a2l1x_2x[] = {GLYPH_x, 0};
+static const glyph_t a2l1x_2y[] = {GLYPH_y, 0};
+static const glyph_t a2l1x_2z[] = {GLYPH_z, 0};
+
+static const glyph_t a2l1x_2zero[] = {GLYPH_zero, 0};
+static const glyph_t a2l1x_2one[] = {GLYPH_one, 0};
+static const glyph_t a2l1x_2two[] = {GLYPH_two, 0};
+static const glyph_t a2l1x_2three[] = {GLYPH_three, 0};
+static const glyph_t a2l1x_2four[] = {GLYPH_four, 0};
+static const glyph_t a2l1x_2five[] = {GLYPH_five, 0};
+static const glyph_t a2l1x_2six[] = {GLYPH_six, 0};
+static const glyph_t a2l1x_2seven[] = {GLYPH_seven, 0};
+static const glyph_t a2l1x_2eight[] = {GLYPH_eight, 0};
+static const glyph_t a2l1x_2nine[] = {GLYPH_nine, 0};
 
 /* These are shared between ascii and latin1. Latin9 uses them as well,
  * but discards the GLYPH_oe and GLYPH_OE tweaks. */
@@ -685,7 +792,159 @@ static const glyphtweak_t a2l1_glyphtweaks[] = {
     { GLYPH_UNI_longst, a2l1x_UNI_longst, WIDTH_AUTO },
     { GLYPH_minus, a2l1x_minus, WIDTH_AUTO },
     { GLYPH_asteriskmath, a2l1x_asteriskmath, WIDTH_AUTO},
-    { GLYPH_UNI_ZWNJ, a2l1x_2emptystring, WIDTH_AUTO}
+    { GLYPH_UNI_ZWNJ, a2l1x_2emptystring, WIDTH_AUTO},
+    { GLYPH_tilde, a2l1x_2asciitilde, WIDTH_AUTO},
+    { GLYPH_similar, a2l1x_2asciitilde, WIDTH_AUTO},
+    { GLYPH_lessequal, a2l1x_lessequal, WIDTH_AUTO},
+    { GLYPH_greaterequal, a2l1x_greaterequal, WIDTH_AUTO},
+
+    { GLYPH_UNI_Amathscript, a2l1x_2A, WIDTH_AUTO },
+    { GLYPH_UNI_Bscript, a2l1x_2B, WIDTH_AUTO },
+    { GLYPH_UNI_Cmathscript, a2l1x_2C, WIDTH_AUTO },
+    { GLYPH_UNI_Dmathscript, a2l1x_2D, WIDTH_AUTO },
+    { GLYPH_UNI_Escript, a2l1x_2E, WIDTH_AUTO },
+    { GLYPH_UNI_Fscript, a2l1x_2F, WIDTH_AUTO },
+    { GLYPH_UNI_Gmathscript, a2l1x_2G, WIDTH_AUTO },
+    { GLYPH_UNI_Hscript, a2l1x_2H, WIDTH_AUTO },
+    { GLYPH_UNI_Iscript, a2l1x_2I, WIDTH_AUTO },
+    { GLYPH_UNI_Jmathscript, a2l1x_2J, WIDTH_AUTO },
+    { GLYPH_UNI_Kmathscript, a2l1x_2K, WIDTH_AUTO },
+    { GLYPH_UNI_Lscript, a2l1x_2L, WIDTH_AUTO },
+    { GLYPH_UNI_Mscript, a2l1x_2M, WIDTH_AUTO },
+    { GLYPH_UNI_Nmathscript, a2l1x_2N, WIDTH_AUTO },
+    { GLYPH_UNI_Omathscript, a2l1x_2O, WIDTH_AUTO },
+    { GLYPH_UNI_Pmathscript, a2l1x_2P, WIDTH_AUTO },
+    { GLYPH_UNI_Qmathscript, a2l1x_2Q, WIDTH_AUTO },
+    { GLYPH_UNI_Rscript, a2l1x_2R, WIDTH_AUTO },
+    { GLYPH_UNI_Smathscript, a2l1x_2S, WIDTH_AUTO },
+    { GLYPH_UNI_Tmathscript, a2l1x_2T, WIDTH_AUTO },
+    { GLYPH_UNI_Umathscript, a2l1x_2U, WIDTH_AUTO },
+    { GLYPH_UNI_Vmathscript, a2l1x_2V, WIDTH_AUTO },
+    { GLYPH_UNI_Wmathscript, a2l1x_2W, WIDTH_AUTO },
+    { GLYPH_UNI_Xmathscript, a2l1x_2X, WIDTH_AUTO },
+    { GLYPH_UNI_Ymathscript, a2l1x_2Y, WIDTH_AUTO },
+    { GLYPH_UNI_Zmathscript, a2l1x_2Z, WIDTH_AUTO },
+
+    { GLYPH_UNI_lscript, a2l1x_2l, WIDTH_AUTO },
+
+    { GLYPH_UNI_Amathbb, a2l1x_2A, WIDTH_AUTO },
+    { GLYPH_UNI_Bmathbb, a2l1x_2B, WIDTH_AUTO },
+    { GLYPH_UNI_Cdblstruck, a2l1x_2C, WIDTH_AUTO },
+    { GLYPH_UNI_Dmathbb, a2l1x_2D, WIDTH_AUTO },
+    { GLYPH_UNI_Emathbb, a2l1x_2E, WIDTH_AUTO },
+    { GLYPH_UNI_Fmathbb, a2l1x_2F, WIDTH_AUTO },
+    { GLYPH_UNI_Gmathbb, a2l1x_2G, WIDTH_AUTO },
+    { GLYPH_UNI_Hdblstruck, a2l1x_2H, WIDTH_AUTO },
+    { GLYPH_UNI_Imathbb, a2l1x_2I, WIDTH_AUTO },
+    { GLYPH_UNI_Jmathbb, a2l1x_2J, WIDTH_AUTO },
+    { GLYPH_UNI_Kmathbb, a2l1x_2K, WIDTH_AUTO },
+    { GLYPH_UNI_Lmathbb, a2l1x_2L, WIDTH_AUTO },
+    { GLYPH_UNI_Mmathbb, a2l1x_2M, WIDTH_AUTO },
+    { GLYPH_UNI_Ndblstruck, a2l1x_2N, WIDTH_AUTO },
+    { GLYPH_UNI_Omathbb, a2l1x_2O, WIDTH_AUTO },
+    { GLYPH_UNI_Pdblstruck, a2l1x_2P, WIDTH_AUTO },
+    { GLYPH_UNI_Qdblstruck, a2l1x_2Q, WIDTH_AUTO },
+    { GLYPH_UNI_Rdblstruck, a2l1x_2R, WIDTH_AUTO },
+    { GLYPH_UNI_Smathbb, a2l1x_2S, WIDTH_AUTO },
+    { GLYPH_UNI_Tmathbb, a2l1x_2T, WIDTH_AUTO },
+    { GLYPH_UNI_Umathbb, a2l1x_2U, WIDTH_AUTO },
+    { GLYPH_UNI_Vmathbb, a2l1x_2V, WIDTH_AUTO },
+    { GLYPH_UNI_Wmathbb, a2l1x_2W, WIDTH_AUTO },
+    { GLYPH_UNI_Xmathbb, a2l1x_2X, WIDTH_AUTO },
+    { GLYPH_UNI_Ymathbb, a2l1x_2Y, WIDTH_AUTO },
+    { GLYPH_UNI_Zdblstruck, a2l1x_2Z, WIDTH_AUTO },
+
+    { GLYPH_UNI_amathbb, a2l1x_2a, WIDTH_AUTO },
+    { GLYPH_UNI_bmathbb, a2l1x_2b, WIDTH_AUTO },
+    { GLYPH_UNI_cmathbb, a2l1x_2c, WIDTH_AUTO },
+    { GLYPH_UNI_dmathbb, a2l1x_2d, WIDTH_AUTO },
+    { GLYPH_UNI_emathbb, a2l1x_2e, WIDTH_AUTO },
+    { GLYPH_UNI_fmathbb, a2l1x_2f, WIDTH_AUTO },
+    { GLYPH_UNI_gmathbb, a2l1x_2g, WIDTH_AUTO },
+    { GLYPH_UNI_hmathbb, a2l1x_2h, WIDTH_AUTO },
+    { GLYPH_UNI_imathbb, a2l1x_2i, WIDTH_AUTO },
+    { GLYPH_UNI_jmathbb, a2l1x_2j, WIDTH_AUTO },
+    { GLYPH_UNI_kmathbb, a2l1x_2k, WIDTH_AUTO },
+    { GLYPH_UNI_lmathbb, a2l1x_2l, WIDTH_AUTO },
+    { GLYPH_UNI_mmathbb, a2l1x_2m, WIDTH_AUTO },
+    { GLYPH_UNI_nmathbb, a2l1x_2n, WIDTH_AUTO },
+    { GLYPH_UNI_omathbb, a2l1x_2o, WIDTH_AUTO },
+    { GLYPH_UNI_pmathbb, a2l1x_2p, WIDTH_AUTO },
+    { GLYPH_UNI_qmathbb, a2l1x_2q, WIDTH_AUTO },
+    { GLYPH_UNI_rmathbb, a2l1x_2r, WIDTH_AUTO },
+    { GLYPH_UNI_smathbb, a2l1x_2s, WIDTH_AUTO },
+    { GLYPH_UNI_tmathbb, a2l1x_2t, WIDTH_AUTO },
+    { GLYPH_UNI_umathbb, a2l1x_2u, WIDTH_AUTO },
+    { GLYPH_UNI_vmathbb, a2l1x_2v, WIDTH_AUTO },
+    { GLYPH_UNI_wmathbb, a2l1x_2w, WIDTH_AUTO },
+    { GLYPH_UNI_xmathbb, a2l1x_2x, WIDTH_AUTO },
+    { GLYPH_UNI_ymathbb, a2l1x_2y, WIDTH_AUTO },
+    { GLYPH_UNI_zmathbb, a2l1x_2z, WIDTH_AUTO },
+
+    { GLYPH_UNI_zeromathbb, a2l1x_2zero, WIDTH_AUTO },
+    { GLYPH_UNI_onemathbb, a2l1x_2one, WIDTH_AUTO },
+    { GLYPH_UNI_twomathbb, a2l1x_2two, WIDTH_AUTO },
+    { GLYPH_UNI_threemathbb, a2l1x_2three, WIDTH_AUTO },
+    { GLYPH_UNI_fourmathbb, a2l1x_2four, WIDTH_AUTO },
+    { GLYPH_UNI_fivemathbb, a2l1x_2five, WIDTH_AUTO },
+    { GLYPH_UNI_sixmathbb, a2l1x_2six, WIDTH_AUTO },
+    { GLYPH_UNI_sevenmathbb, a2l1x_2seven, WIDTH_AUTO },
+    { GLYPH_UNI_eightmathbb, a2l1x_2eight, WIDTH_AUTO },
+    { GLYPH_UNI_ninemathbb, a2l1x_2nine, WIDTH_AUTO },
+
+    { GLYPH_UNI_Amathfrak, a2l1x_2A, WIDTH_AUTO },
+    { GLYPH_UNI_Bmathfrak, a2l1x_2B, WIDTH_AUTO },
+    { GLYPH_UNI_Cfraktur, a2l1x_2C, WIDTH_AUTO },
+    { GLYPH_UNI_Dmathfrak, a2l1x_2D, WIDTH_AUTO },
+    { GLYPH_UNI_Emathfrak, a2l1x_2E, WIDTH_AUTO },
+    { GLYPH_UNI_Fmathfrak, a2l1x_2F, WIDTH_AUTO },
+    { GLYPH_UNI_Gmathfrak, a2l1x_2G, WIDTH_AUTO },
+    { GLYPH_UNI_Hfraktur, a2l1x_2H, WIDTH_AUTO },
+    { GLYPH_Ifraktur, a2l1x_2I, WIDTH_AUTO },
+    { GLYPH_UNI_Jmathfrak, a2l1x_2J, WIDTH_AUTO },
+    { GLYPH_UNI_Kmathfrak, a2l1x_2K, WIDTH_AUTO },
+    { GLYPH_UNI_Lmathfrak, a2l1x_2L, WIDTH_AUTO },
+    { GLYPH_UNI_Mmathfrak, a2l1x_2M, WIDTH_AUTO },
+    { GLYPH_UNI_Nmathfrak, a2l1x_2N, WIDTH_AUTO },
+    { GLYPH_UNI_Omathfrak, a2l1x_2O, WIDTH_AUTO },
+    { GLYPH_UNI_Pmathfrak, a2l1x_2P, WIDTH_AUTO },
+    { GLYPH_UNI_Qmathfrak, a2l1x_2Q, WIDTH_AUTO },
+    { GLYPH_Rfraktur, a2l1x_2R, WIDTH_AUTO },
+    { GLYPH_UNI_Smathfrak, a2l1x_2S, WIDTH_AUTO },
+    { GLYPH_UNI_Tmathfrak, a2l1x_2T, WIDTH_AUTO },
+    { GLYPH_UNI_Umathfrak, a2l1x_2U, WIDTH_AUTO },
+    { GLYPH_UNI_Vmathfrak, a2l1x_2V, WIDTH_AUTO },
+    { GLYPH_UNI_Wmathfrak, a2l1x_2W, WIDTH_AUTO },
+    { GLYPH_UNI_Xmathfrak, a2l1x_2X, WIDTH_AUTO },
+    { GLYPH_UNI_Ymathfrak, a2l1x_2Y, WIDTH_AUTO },
+    { GLYPH_UNI_Zfraktur, a2l1x_2Z, WIDTH_AUTO },
+
+    { GLYPH_UNI_amathfrak, a2l1x_2a, WIDTH_AUTO },
+    { GLYPH_UNI_bmathfrak, a2l1x_2b, WIDTH_AUTO },
+    { GLYPH_UNI_cmathfrak, a2l1x_2c, WIDTH_AUTO },
+    { GLYPH_UNI_dmathfrak, a2l1x_2d, WIDTH_AUTO },
+    { GLYPH_UNI_emathfrak, a2l1x_2e, WIDTH_AUTO },
+    { GLYPH_UNI_fmathfrak, a2l1x_2f, WIDTH_AUTO },
+    { GLYPH_UNI_gmathfrak, a2l1x_2g, WIDTH_AUTO },
+    { GLYPH_UNI_hmathfrak, a2l1x_2h, WIDTH_AUTO },
+    { GLYPH_UNI_imathfrak, a2l1x_2i, WIDTH_AUTO },
+    { GLYPH_UNI_jmathfrak, a2l1x_2j, WIDTH_AUTO },
+    { GLYPH_UNI_kmathfrak, a2l1x_2k, WIDTH_AUTO },
+    { GLYPH_UNI_lmathfrak, a2l1x_2l, WIDTH_AUTO },
+    { GLYPH_UNI_mmathfrak, a2l1x_2m, WIDTH_AUTO },
+    { GLYPH_UNI_nmathfrak, a2l1x_2n, WIDTH_AUTO },
+    { GLYPH_UNI_omathfrak, a2l1x_2o, WIDTH_AUTO },
+    { GLYPH_UNI_pmathfrak, a2l1x_2p, WIDTH_AUTO },
+    { GLYPH_UNI_qmathfrak, a2l1x_2q, WIDTH_AUTO },
+    { GLYPH_UNI_rmathfrak, a2l1x_2r, WIDTH_AUTO },
+    { GLYPH_UNI_smathfrak, a2l1x_2s, WIDTH_AUTO },
+    { GLYPH_UNI_tmathfrak, a2l1x_2t, WIDTH_AUTO },
+    { GLYPH_UNI_umathfrak, a2l1x_2u, WIDTH_AUTO },
+    { GLYPH_UNI_vmathfrak, a2l1x_2v, WIDTH_AUTO },
+    { GLYPH_UNI_wmathfrak, a2l1x_2w, WIDTH_AUTO },
+    { GLYPH_UNI_xmathfrak, a2l1x_2x, WIDTH_AUTO },
+    { GLYPH_UNI_ymathfrak, a2l1x_2y, WIDTH_AUTO },
+    { GLYPH_UNI_zmathfrak, a2l1x_2z, WIDTH_AUTO }
 };
 
 static void ascii_init(void)
@@ -929,7 +1188,7 @@ static void latin9_init(void)
  * implementations of encoding independent stuff
  ***********************************************************************/
 
-static sparp_t glyphwidth_sparp;
+static spars32_t glyphwidth_spars32;
 static sparp_t glyphxlat_sparp;
 
 void outenc_init(void)
@@ -940,7 +1199,7 @@ void outenc_init(void)
     
     sparp_init(&accenting_sparp);
     vlist_init(&accenting_vlist);
-    sparp_init(&glyphwidth_sparp);
+    spars32_init(&glyphwidth_spars32, 1);
     sparp_init(&glyphxlat_sparp);
 
     outenc_register_glyphtweaks(
@@ -979,22 +1238,27 @@ void outenc_init(void)
 
 int outenc_get_width(const linebuf_t * l)
 {
-    glyph_t * p, * q;
+    glyph_t * p;
     int w = 0;
-    int cl;
-    
+
     pmesg(50, "BEGIN outenc_get_width\n");
-    
+
     p = l->gstring;
     while(*p != 0) {
-    	q = p + 1;
-	while(glyph_get_hint(*q) & GH_COMBINING) ++q;
-	    /* will stop by itself on 0 termination */
-	cl = q - p;
-	if(cl == 1) w += outenc_glyph_width(*p);
-	else w += outenc_composition_width(p, cl);
-	p = q;
+    	if(glyph_get_hint(*(p + 1)) & GH_COMBINING) {
+	    /* grok the sequence of combining diacritics after *p */
+    	    glyph_t * q = p + 2;
+	    while(glyph_get_hint(*q) & GH_COMBINING) ++q;
+	    	/* will stop by itself on 0 termination */
+	    w += outenc_composition_width(p, q - p);
+	    p = q;
+	}
+	else {
+	    w += outenc_glyph_width(*p);
+	    ++p;
+	}
     }
+
     pmesg(50, "END outenc_get_width\n");
     return w;
 }
@@ -1003,27 +1267,31 @@ int outenc_get_width(const linebuf_t * l)
 void outenc_write(FILE * f, const linebuf_t * l)
 {
     linebuf_t xl;
-    glyph_t * p, * q;
-    int cl;
-    
+    glyph_t * p;
+
     pmesg(50, "BEGIN outenc_write\n");
-    
+
     linebuf_init(&xl, 2*l->size_curr);
-    
+
     p = l->gstring;
     while(*p != 0) {
-    	q = p + 1;
-	while(glyph_get_hint(*q) & GH_COMBINING) ++q;
-	    /* will stop by itself on 0 termination */
-	cl = q - p;
-	if(cl == 1) outenc_xlat_glyph(*p, &xl);
-	else outenc_xlat_composition(p, cl, &xl);
-	p = q;
+    	if(glyph_get_hint(*(p + 1)) & GH_COMBINING) {
+	    /* grok the sequence of combining diacritics after *p */
+    	    glyph_t * q = p + 2;
+	    while(glyph_get_hint(*q) & GH_COMBINING) ++q;
+	    	/* will stop by itself on 0 termination */
+	    outenc_xlat_composition(p, q - p, &xl);
+	    p = q;
+	}
+	else {
+	    outenc_xlat_glyph(*p, &xl);
+	    ++p;
+	}
     }
-    
+
     p = xl.gstring;
     while(*p != 0) outenc_write_raw_glyph(f, *p++);
-    
+
     linebuf_done(&xl);
 
     pmesg(50, "END outenc_write\n");
@@ -1031,10 +1299,7 @@ void outenc_write(FILE * f, const linebuf_t * l)
 
 static int outenc_glyph_width(glyph_t g)
 {
-    int * p;
-    
-    p = sparp_read(&glyphwidth_sparp, g);
-    return(p == NULL ? 1 : *p);
+    return (int) spars32_read(&glyphwidth_spars32, g);
 }
 
 static void outenc_xlat_glyph(glyph_t g, linebuf_t * l)
@@ -1050,18 +1315,17 @@ static void outenc_xlat_glyph(glyph_t g, linebuf_t * l)
 static void outenc_register_glyphtweaks(const glyphtweak_t tweaks[], int length)
 {
     int i;
-    int * pw;
     linebuf_t * px;
     glyphtweak_t tweak;
         
-    pw = malloc(length * sizeof(int));
-    if(pw == NULL) enomem();
     px = malloc(length * sizeof(linebuf_t));
     if(px == NULL) enomem();
-    /* Note we "leak memory" here since we leave pw, px allocated but
+    /* Note we "leak memory" here since we leave px allocated but
      * forget them on returning from this function. This is OK since
      * the allocated memory is used throughout the whole lifetime of
      * the process.
+     *
+     * FIXME: this has to stop if we want to become a library.
      */
      
     for(i=0; i < length; ++i) {
@@ -1073,8 +1337,7 @@ static void outenc_register_glyphtweaks(const glyphtweak_t tweaks[], int length)
 	    sparp_write(&glyphxlat_sparp, tweak.glyph, px+i);
 	    if(tweak.width == WIDTH_AUTO) tweak.width = px[i].size_curr;
 	}
-	pw[i] = tweak.width;
-	sparp_write(&glyphwidth_sparp, tweak.glyph, pw+i);
+	spars32_write(&glyphwidth_spars32, tweak.glyph, tweak.width);
     }
 
 }
@@ -1084,8 +1347,8 @@ static void outenc_unregister_glyphtweak(glyph_t g)
     if(sparp_read(&glyphxlat_sparp, g) != NULL) {
     	sparp_write(&glyphxlat_sparp, g, NULL);
     }
-    if(sparp_read(&glyphwidth_sparp, g) != NULL) {
-    	sparp_write(&glyphwidth_sparp, g, NULL);
+    if(spars32_read(&glyphwidth_spars32, g) != 1) {
+    	spars32_write(&glyphwidth_spars32, g, 1);
     }
 }
 
