@@ -1,6 +1,6 @@
 /* catdvi - get text from DVI files
    Copyright (C) 1999 Antti-Juhani Kaijanaho <gaia@iki.fi>
-   Copyright (C) 2000-01 Bjoern Brill <brill@fs.math.uni-frankfurt.de>
+   Copyright (C) 2000-02 Bjoern Brill <brill@fs.math.uni-frankfurt.de>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -19,57 +19,10 @@
 
 #include <assert.h>
 #include <stdlib.h>
-#include "fontinfo.h"
-#include "page.h"
-#include "outenc.h"
+#include "page2.h"
 #include "util.h"
 #include "glyphops.h"
-#include "linebuf.h"
-#include "density.h"
-#include <errno.h>
-
-/* A page is modeled as an ordered doubly-linked list of boxes (x, y,
-   glyph), where the ordering is left-to-right top-to-bottom. */
-
-enum box_flag_t {
-    BF_SWIMMING =   1 << 0,
-    	/* (x, y) != (left border, baseline). We need to
-    	 * move this one to the "logically right" place.
-	 */
-    BF_HAS_AXIS =   1 << 1,
-    	/* box_t.axis is valid, i.e. we know where the math
-    	 * axis passes through this box.
-	 */
-    BF_ON_AXIS =    1 << 2,
-    	/* This box is centered on the math axis (e.g. the
-    	 * big operators). Implies BF_SWIMMING.
-    	 */
-    BF_RADICAL =    1 << 3,
-    	/* This is a radical sign (and hence "hanging down"
-    	 * from the y coordinate). Implies BF_SWIMMING.
-	 */
-    BF_DIACRITIC =  1 << 4 
-    	/* This is a diacritical mark. Often its baseline is vertically
-	 * displaced against the baseline of the glyph it should accent
-    	 * because the combination looks better this way).
-	 * Implies BF_SWIMMING.
-	 */
-};
-   
-struct box_t {
-        sint32 x;
-        sint32 y;
-	sint32 axis;
-	
-        sint32 width;
-        sint32 height;
-	sint32 depth;
-	sint32 axis_height; /* The TeX font parameter. <= 0 if unknown. */
-	
-        glyph_t glyph;
-        font_t font;
-	enum box_flag_t flags;
-};
+#include "layout.h"
 
 /* Return value like with strcmp and friends. */
 static int compare_boxes(struct box_t a, struct box_t b)
@@ -100,12 +53,6 @@ static int point_in_box(const struct point_t * point, const struct box_t * box)
     );
 }
 
-struct list_node_t {
-        struct box_t b;
-        struct list_node_t * prev;
-        struct list_node_t * next;
-};
-
 /* Set to 1 when catdvi is invoked with the --sequential option */
 int page_sequential = 0;
 
@@ -115,14 +62,9 @@ int page_list_numbers = 0;
 /* List support variables.  Note that most insertions happen near the
    previous insertion, so it's beneficial to keep a pointer to the
    last inserted node. */
-static struct list_node_t * list_head;
-static struct list_node_t * list_tail;
-static struct list_node_t * list_latest; /* Node that was inserted last. */
-
-/* left and right margins - set by page_end(), also used by
- * decide_space_count() and page_print().
- */
-static sint32 left_margin, right_margin;
+struct list_node_t * list_head;
+struct list_node_t * list_tail;
+struct list_node_t * list_latest; /* Node that was inserted last. */
 
 /* flags to avoid the page_adjust_*() loops over the whole page when
  * there's nothing to adjust.
@@ -190,6 +132,15 @@ static void bubble(void)
         NOTREACHED;
 }
 
+
+/* move some random node into its correct position in the list */
+static void bubble_node(struct list_node_t * p)
+{
+    list_latest = p;
+    bubble();
+}
+
+
 static void insert_list(struct box_t box)
 {
         struct list_node_t * new;
@@ -304,93 +255,48 @@ void page_set_glyph(
     sint32 x, sint32 y
 )
 {
-        struct box_t b;
+    struct box_t b;
+    enum glyph_hint_t hint;
 
-        b.x = x;
-        b.y = y;
-        b.width = width;
-        b.height = height;
-	b.depth = depth;
-        b.glyph = glyph;
-        b.font = font;
-	b.axis_height = axis_height;
-	b.flags = 0;
-	b.axis = 0;
-	
-    	if(glyph_get_hint(glyph) & GH_DIACRITIC) {
-	    b.flags |= BF_SWIMMING | BF_DIACRITIC ;
-	    page_has_diacritics = 1;
-	}
-	    
-	if(glyph_get_hint(glyph) & GH_EXTENSIBLE_RECIPE) {
-	    b.flags |= BF_SWIMMING;
-	    page_has_texmext = 1;
-	}
-	else if(glyph_get_hint(glyph) & GH_ON_AXIS) {
-	    b.flags |= BF_SWIMMING | BF_ON_AXIS | BF_HAS_AXIS ;
-	    b.axis = y + (-height + depth)/2;
-	    page_has_texmext = 1;
-	}
-	else if(glyph_get_hint(glyph) & GH_RADICAL) {
-	    b.flags |= BF_SWIMMING | BF_RADICAL ;
-	    page_has_radicals = 1;
-	}
-	else if(axis_height > 0) {
-	    b.flags |= BF_HAS_AXIS;
-	    b.axis = y - axis_height;
-	}
+    b.x = x;
+    b.y = y;
+    b.width = width;
+    b.height = height;
+    b.depth = depth;
+    b.glyph = glyph;
+    b.font = font;
+    b.axis_height = axis_height;
+    b.flags = 0;
+    b.axis = 0;
 
-        insert_list(b);
+    hint = glyph_get_hint(glyph);
+
+    if(hint & GH_DIACRITIC) {
+	b.flags |= BF_SWIMMING | BF_DIACRITIC ;
+	page_has_diacritics = 1;
+    }
+
+    if(hint & GH_EXTENSIBLE_RECIPE) {
+	b.flags |= BF_SWIMMING;
+	page_has_texmext = 1;
+    }
+    else if(hint & GH_ON_AXIS) {
+	b.flags |= BF_SWIMMING | BF_ON_AXIS | BF_HAS_AXIS ;
+	b.axis = y + (-height + depth)/2;
+	page_has_texmext = 1;
+    }
+    else if(hint & GH_RADICAL) {
+	b.flags |= BF_SWIMMING | BF_RADICAL ;
+	page_has_radicals = 1;
+    }
+    else if(axis_height > 0) {
+	b.flags |= BF_HAS_AXIS;
+	b.axis = y - axis_height;
+    }
+
+    insert_list(b);
 }
 
-static int decide_space_count(
-    struct list_node_t * prev_p,
-    struct list_node_t * p
-)
-{
-        font_t prev_font, curr_font;
-        sint32 prev_x, prev_width;
-        sint32 curr_x;
-        sint32 delta;
-        int (*f2spc)(sint32, sint32) = font_w_to_space;
-
-        assert((prev_p != 0) || (p != 0));
-
-
-        /* If we are at the start of a new line, we don't use the previous
-	 * box. Otherwise, the decision depends on the previous box too.
-	 */
-	if (prev_p && p && (prev_p->b.y != p->b.y)) prev_p = NULL;
-        if (prev_p) {
-                prev_font = prev_p->b.font;
-                prev_x = prev_p->b.x;
-                prev_width = prev_p->b.width;
-        } else {
-                prev_font = p->b.font;
-                prev_x = left_margin;
-                prev_width = 0;
-        }
-
-    	/* To estimate the amount of whitespace that required to
-	 * fill the space from the end of the line to the right margin,
-	 * we're called with p == NULL
-	 */
-        if (p) {
-        	curr_font = p->b.font;
-        	curr_x = p->b.x;
-        } else {
-        	curr_font = prev_font;
-        	curr_x = right_margin;
-        }
-	
-        delta = curr_x - (prev_x + prev_width);
-
-        return (1 + f2spc(curr_font, delta) + f2spc(prev_font, delta)) / 2;
-}
-
-static void page_end_sequential(void);
-
-static void page_adjust_diacritics(void);
 
 static void page_pair_accenting(
     struct list_node_t * pbase,
@@ -398,378 +304,37 @@ static void page_pair_accenting(
     int direction
 );
 
-static void page_adjust_texmext(void);
-static void page_adjust_radicals(void);
-
-static void page_print(
-    struct linebuf_t lines[],
-    sint32 lines_ypos[],
-    int nlines,
-    struct scdf_t * col_density
-);
-
 void page_end(void)
 {
-        struct list_node_t * p, * prev_p;
-        sint32 prev_y, height = 1;
-	sint32 prev_tab, this_tab;
-        int linecount = 0;
-	int last_word_width = 0;
-        int space_count, space_reserve;
-	struct linebuf_t * lbs, * lb;
-	struct linebuf_t last_word;
-	struct scdf_t col_density;
-	sint32 * lines_ypos, * this_ypos;
+    pmesg(50, "BEGIN page_end\n");
 
-        pmesg(50, "BEGIN page_end\n");
-	
-    	if (
-	    pageref_cmp(&page_start_output, &current_page) > 0 ||
-	    pageref_cmp(&page_last_output, &current_page) < 0
-	) {
-	    if(msglevel >= 80) {
-	    	pmesg(80, "skipping page by user request:\n  ");
-		pageref_print(&current_page, stderr);
-	    }
-            pmesg(50, "END page_end\n");
-	    return;
+    if (
+	pageref_cmp(&page_start_output, &current_page) > 0 ||
+	pageref_cmp(&page_last_output, &current_page) < 0
+    ) {
+	if(msglevel >= 80) {
+	    pmesg(80, "skipping page by user request:\n  ");
+	    pageref_print(&current_page, stderr);
 	}
+	pmesg(50, "END page_end\n");
+	return;
+    }
 
-    	if(page_list_numbers) {
-	    pageref_print(&current_page, stdout);
-            pmesg(50, "END page_end\n");
-	    return;
-	}
-	
-    	page_adjust_diacritics();
-	
-    	if(page_sequential) {
-	    page_end_sequential();
-            pmesg(50, "END page_end\n");
-	    return;
-	};
-	
-	/* Life is much easier if we don't have to care for the case
-	 * of an empty page all the time.
-	 */
-    	if(list_head == NULL) {
-            puts("\f\n");
-	    pmesg(80, "This page is empty.\n");
-            pmesg(50, "END page_end\n");
-	    return;
-	}
+    if(page_list_numbers) {
+	pageref_print(&current_page, stdout);
+	pmesg(50, "END page_end\n");
+	return;
+    }
 
-    	page_adjust_texmext();
-    	page_adjust_radicals();
+    if(page_sequential) page_print_sequential();
+    else page_print_formatted();
+    
+    pmesg(50, "END page_end\n");
+    return;
 
-        /* Count lines, find left and right margins */
-	left_margin = SINT32_MAX;
-	right_margin = SINT32_MIN;
-	    /* highest and lowest value guaranteed to fit into an sint32 */
-	prev_y = list_head->b.y;
-        for (p = list_head; p != 0; p = p->next) {
-	    	left_margin = min(left_margin, p->b.x);
-		right_margin = max(right_margin, p->b.x + p->b.width); 
-                if (p->b.y > prev_y) {
-                        prev_y = p->b.y;
-                        linecount++;
-                }
-        }
-	++linecount; /* The last line hasn't been counted */
-	pmesg(80, "left margin: %ld, right margin: %ld.\n", \
-	    left_margin, right_margin);
-
-    	/* col_dens measures the column density at every point between
-	 * left and right margin which is required to do proper positioning
-	 * of every tab stop (see below).
-	 */
-    	scdf_init(&col_density, left_margin, right_margin, 0);
-	
-	/* The text contained in the page is stored in an array of linebufs,
-	 * one for each line. Printout is done in one go by page_print()
-	 * after we have parsed the complete page and collected the
-	 * neccessary layout data.
-	 */
-	lbs = malloc(linecount * sizeof(struct linebuf_t));
-	if(lbs == NULL) enomem();
-	lb = lbs;
-	linebuf_init(lb, 0);
-	
-    	/* The y position of every line is stored in an array so that
-	 * page_print() can do some vertical formatting.
-	 */
-	lines_ypos = malloc(linecount * sizeof(sint32));
-	if(lines_ypos == NULL) enomem();
-	this_ypos = lines_ypos;
-	*this_ypos = list_head->b.y;
-	
-	/* We collect the current run of not space-separated glyphs
-	 * in last_word
-	 */
-	linebuf_init(&last_word, 0);
-	
-	prev_y = list_head->b.y;
-	prev_tab = left_margin;
-	prev_p = NULL;
-        for (p = list_head; p != 0; prev_p = p, p = p->next) {
-
-                pmesg(80, "node: X = %li, Y = %li, glyph = %li (%c), font = %li\n",
-                      p->b.x, p->b.y, p->b.glyph, (unsigned char) p->b.glyph, p->b.font);
-
-                if (p->b.height != 0) height = p->b.height;
-
-                if (p->b.y > prev_y) {
-                        pmesg(80, "end of line\n");
-			linebuf_append(lb, &last_word);
-			linebuf_init(++lb, 0);
-			*(++this_ypos) = p->b.y;
-                        prev_y = p->b.y;
-			
-			space_count = decide_space_count(prev_p, NULL);
-			space_reserve = (space_count + 1) / 2;
-			last_word_width = outenc_get_width(&last_word);
-			scdf_force_min_integral(
-			    &col_density,
-			    prev_tab,
-			    right_margin,
-			    last_word_width + space_reserve
-			);
-			scdf_normalize(&col_density);
-			linebuf_clear(&last_word);
-			prev_tab = left_margin;
-                }
-
-                space_count = decide_space_count(prev_p, p);
-                
-		
-    	    	if((space_count > 0) && (p->b.x != prev_tab)) {
-		    /* Word beginnings are used as "tab stops" where
-		     * physical position and printout column position
-		     * get synchronized.
-		     * We make sure there are enough columns to hold
-		     * all characters since the last tab stop by requiring
-		     * that the column density in between those tabs
-		     * be high enough.
-		     */
-		    last_word_width = outenc_get_width(&last_word);
-		    this_tab = p->b.x;
-		    space_reserve = (space_count + 1) / 2;
-			    /* This formula is a compromise. On one hand,
-			     * we don't want to waste too much space with
-			     * spaces :) but on the other, we'd like if
-			     * larger portions of whitespace (e.g between
-			     * columns) would show up in the printout as well.
-			     */
-		    pmesg(
-		    	80,
-			"prev_tab is %ld, this_tab is %ld, space_count %d\n",
-			prev_tab,
-			this_tab,
-			space_count
-		    );
-		    scdf_force_min_integral(
-		    	&col_density,
-			prev_tab,
-			this_tab,
-			last_word_width + space_reserve
-		    );
-		    prev_tab = this_tab;
-
-    	    	    linebuf_append(lb, &last_word);
-		    linebuf_clear(&last_word);
-
-    		    /* insert one space (just to make really sure we'll
-		     * see one in the printout) and a tab stop in the
-		     * line buffer
-		     */
-    	    	    linebuf_putg(lb, ' ');
-		    linebuf_putg(lb, '\t');
-		    linebuf_putg(lb, last_word_width + 1);
-		    	/* +1 for the space */
-		    linebuf_putg(lb, this_tab);
-		}
-
-                linebuf_putg(&last_word, p->b.glyph);
-        }
-	
-    	/* End of page. We still have to end the last line */
-    	space_count = decide_space_count(prev_p, NULL);
-	space_reserve = (space_count + 1) / 2;
-    	last_word_width = outenc_get_width(&last_word);
-	scdf_force_min_integral(
-	    &col_density,
-	    prev_tab,
-	    right_margin,
-	    last_word_width + space_reserve
-	);
-	scdf_normalize(&col_density);
-	
-    	linebuf_append(lb, &last_word);
-
-	/* debugging blurb */
-    	if(msglevel >= 100) {
-	    scdf_dump(&col_density);
-    	    fprintf(
-		stderr,
-		"Estimated number of required columns is %10.9g\n",
-		scdf_integral(&col_density, left_margin, right_margin)
-	    );
-	}
-
-    	page_print(lbs, lines_ypos, linecount, &col_density);
-        puts("\f\n");
-    	/* clean up */
-	linebuf_done(&last_word);
-	for(lb = lbs; lb - lbs < linecount ; ++lb) linebuf_done(lb);
-	free(lbs);
-    	scdf_done(&col_density);
-	    
-        pmesg(50, "END page_end\n");
 }
 
-static void page_print(
-    struct linebuf_t lines[],
-    sint32 lines_ypos[],
-    int nlines,
-    struct scdf_t * col_density
-)
-{
-    struct linebuf_t * curr_line;
-    struct linebuf_t formatted;
-    int col, ncols;
-    sint32 pos;
-    sint32 * colpos;
-    int i, j;
-    int nlfs;
-    float line_density;
-    
-    pmesg(50, "BEGIN page_print\n");
-    
-    /* Find the positions of the printout columns */
-    ncols = 3 + 2 * (int) scdf_integral(col_density, left_margin, right_margin);
-    	/*
-	 * +1 for the left magin, +1 for the fractional part, +1 since we'll
-	 * have to look one cloumn to the right even from the last real column
-	 * at the right margin and *2 to be on the very safe side wrt extra
-	 * columns produced by rounding errors in the column-cutting algorithm.
-	 *
-	 * I haven't seen any extra columns yet but their number should be
-	 * proportional to the really required number of columns.
-	 */
-    colpos = malloc(ncols * sizeof(sint32));
-    if(colpos == NULL) enomem();
-    
-    pos = colpos[0] = left_margin;
-    col = 1;
-    while(pos < right_margin) {
-    	errno = 0;
-	pos = scdf_solve_integral_for_x1(col_density, pos, 1);
-	if(errno == EDOM) pos = right_margin;
-    	    /* EDOM means not enough density right from pos to
-	     * give integral 1, i.e. we're nearly at the right margin.
-	     */
-	colpos[col++] = pos;
-    }
-    for(j = col; j < ncols; ++j) colpos[col] = right_margin;
-    	/* The rest gets filled with position = right margin to avoid
-	 * bombing out if rounding or other problems lead to a few
-	 * more printout columns than expected. */
-    ncols = col;
-    pmesg(80, "ncols is %d\n", ncols);
-
-    line_density = ((float) nlines) / (lines_ypos[nlines-1] - lines_ypos[0]);
-    
-    /* hardcore debugging stuff */
-    if(msglevel >= 100) {
-    	fputs("Dumping column positions:\n", stderr);
-	for(j = 0; j < ncols - 1 ; ++j) {
-    	    fprintf(
-		stderr,
-		"  col %3d pos = %10ld integral to next col = %10.9g\n",
-		j,
-		colpos[j],
-		scdf_integral(col_density, colpos[j], colpos[j+1])
-    	    );
-	}
-    }
-
-    /* now reformat the lines and print them out */
-    linebuf_init(&formatted, 0);
-
-    for(i = 0; i < nlines; ++i) {
-	
-    	curr_line = lines + i;
-	linebuf_clear(&formatted);
-	col = 0;
-	for(j = 0; j < curr_line->size_curr; ) {
-	    glyph_t g;
-	    
-	    if(col >= ncols) {
-	    	pmesg(10, "Oops: col = %d >= ncols = %d\n", col, ncols);
-	    }
-	    
-	    g = curr_line->gstring[j++];
-	    if(g == '\t') {
-	    	col += curr_line->gstring[j++];
-	    	pos = curr_line->gstring[j++];
-		pmesg(80, "col %3d: tab to pos %10ld.\n", col, pos);
-		if(pos < colpos[col]) {
-	    	    pmesg(10,
-		    "Oops: tab to %10ld < pos = %10ld, difference %10ld\n",
-		    pos,
-		    colpos[col],
-		    colpos[col] - pos);
-		}
-	    	while(colpos[col+1] < pos) {
-		    linebuf_putg(&formatted, ' ');
-		    pmesg(
-		    	80,
-			"  Skipping col %3d at pos %10ld\n",
-			col,
-			colpos[col]
-		    );
-		    ++col;
-		}
-		pmesg(
-		    80,
-		    "col %3d: pos = %10ld, next pos = %10ld\n\n",
-		    col,
-		    colpos[col],
-		    colpos[col+1]
-		);
-	    }
-	    else {
-	    	linebuf_putg(&formatted, g);
-		pmesg(
-		    80,
-		    "col >= %3d: putting glyph U+%04lX [%c]\n",
-		    col,
-		    g,
-		    (unsigned char) g &0xff
-		);
-	    }
-	}
-	
-	outenc_write(stdout, &formatted);
-
-	/* Approximate vertical distance to next line by line feeds */
-	if(i+1 < nlines) {
-	    nlfs = (int) ((lines_ypos[i+1] - lines_ypos[i]) * line_density
-	    	+ 0.5);
-	    nlfs = max(nlfs, 1);
-	}
-	else nlfs = 1;
-	pmesg(80, "\nputting %3d linefeeds\n", nlfs);
-        for (j = 0; j < nlfs; ++j) putc('\n', stdout);
-    }
-    	
-    linebuf_done(&formatted);
-    free(colpos);
-    pmesg(50, "END page_print\n");
-}
-
-
-static void page_adjust_diacritics(void)
+void page_adjust_diacritics(void)
 {
     int trouble = 0;
     int direction = 0;
@@ -937,107 +502,14 @@ static void page_pair_accenting(
 	    pdiacritic->b.height + pbase->b.y - pdiacritic->b.y;
 	pdiacritic->b.y = pbase->b.y;
 	pdiacritic->b.flags &= ~(BF_SWIMMING | BF_DIACRITIC);
-	if(!page_sequential) {
-	    /* keep the list ordered */
-	    list_latest = pdiacritic;
-	    	/* hope this won't fsck up. Better change bubble()s
-		 * parameter list instead */
-	    bubble();
-	}
+
+	/* keep the list ordered */
+	if(!page_sequential) bubble_node(pdiacritic);
     }
 
 }
 
 
-static void page_end_sequential(void)
-{
-    struct list_node_t * p;
-    struct box_t prev_box, curr_box;
-
-    sint32 delta_x = 0, delta_y = 0;
-    sint32 dist_x = 0;
-    sint32 epsilon_x = 0, epsilon_y = 0;
-
-    /* begin-of-page flag */
-    int bop = 1;
-    /* begin-of-line flag */
-    int bol = 1;
-        
-    int spaces, spaces_by_prev, spaces_by_curr;
-    struct linebuf_t lb;    
-
-
-    pmesg(50, "BEGIN page_end_sequential\n");
-
-    linebuf_init(&lb, 0);
-    prev_box.width = 1;
-    prev_box.height = 1;
-    
-    for (p = list_head; p != 0; p = p->next) {
-
-        pmesg(80, "node: X = %li, Y = %li, glyph = %li (%c), font = %li\n",
-              p->b.x, p->b.y, p->b.glyph, (unsigned char) p->b.glyph, p->b.font);
-
-    	curr_box = p->b;
-        if (curr_box.width == 0) curr_box.width = prev_box.width;
-        if (curr_box.height == 0) curr_box.height = prev_box.height;
-
-	if (!bop) {
-    	    delta_x = curr_box.x - prev_box.x;
-	    delta_y = curr_box.y - prev_box.y;
-	    dist_x = delta_x - prev_box.width;
-
-    	    /* (arithmetic mean of current and previous) / 20 */
-	    epsilon_x = (curr_box.width + prev_box.width + 39) / 40;
-	    epsilon_y = (curr_box.height + prev_box.height + 39) / 40;
-
-            /* check for new line */
-	    if (
-		/* new line */
-		delta_y >= 22*epsilon_y ||
-		/* new column */
-		delta_y <= -60*epsilon_y ||
-		/* weird step back */
-		dist_x <= -100*epsilon_x
-	    ) {
-        	pmesg(80, "end of line\n");
-		outenc_write(stdout, &lb);
-		linebuf_clear(&lb);
-		puts("");
-		bol = 1;
-            }
-	    
-	    /* check for word breaks, but not at the beginning of a line */
-	    if ((dist_x > 0) && (!bol)) {
-		spaces_by_prev =
-		    font_w_to_space(prev_box.font, dist_x);
-		spaces_by_curr =
-		    font_w_to_space(curr_box.font, dist_x);
-    		spaces = (spaces_by_prev + spaces_by_curr + 1) / 2;
-
-        	if (spaces > 0) {
-                    pmesg(80, "setting space\n");
-                    linebuf_putg(&lb, ' ');
-        	}
-	    }
-		
-	} /* end if (!bop) */
-
-        linebuf_putg(&lb, curr_box.glyph);
-	prev_box = curr_box;
-	bop = 0;
-	bol = 0;
-    }
-    /* flush line buffer and end page */
-    outenc_write(stdout, &lb);
-    puts("\n\f\n");
-    
-    linebuf_done(&lb);
-    
-    pmesg(50, "END page_end_sequential\n");
-}
-
-
 /************************************************************************
  * the hairy texmext stuff
  ************************************************************************/
@@ -1382,7 +854,7 @@ static struct list_node_t * page_guess_reference_box(
 }
 
 
-static void page_adjust_texmext(void)
+void page_adjust_texmext(void)
 {
     struct list_node_t * p, * s, * ref;
     sint32 new_y, delta;
@@ -1438,10 +910,7 @@ static void page_adjust_texmext(void)
 	p->b.axis_height = p->b.y - p->b.axis;
 	p->b.flags &= ~(BF_SWIMMING | BF_ON_AXIS);
 	
-	if(!page_sequential) {
-	    list_latest = p;
-	    bubble();
-	}
+	if(!page_sequential) bubble_node(p);
 
     }
 
@@ -1489,7 +958,7 @@ static struct list_node_t * page_find_lowest_box(struct searchinfo_t * quest)
  * the bottom tip ought to be) or, if there's no radicant, to the y position
  * of the bottom tip.
  */
-static void page_adjust_radicals(void)
+void page_adjust_radicals(void)
 {
     struct list_node_t * p, * s, * lowest;
     sint32 new_y, delta;
@@ -1561,10 +1030,7 @@ static void page_adjust_radicals(void)
 	p->b.depth -= delta;
 	p->b.flags &= ~(BF_SWIMMING | BF_RADICAL);
 	
-	if(!page_sequential) {
-	    list_latest = p;
-	    bubble();
-	}
+	if(!page_sequential) bubble_node(p);
 
     }
 
